//
//  TetrisBoardView.swift
//  Tetris
//
//  Created by Santiago Pavón Gómez on 9/8/16.
//  Copyright © 2016 UPM. All rights reserved.
//

import UIKit

protocol BoardViewDataSource: class {
    
    /// Preguntar al Data Source cuantas filas tiene el tablero a pintar.
    func numberOfRows(in boardView: BoardView) -> Int
    
    /// Preguntar al Data Source cuantas columnas tiene el tablero a pintar.
    func numberOfColumns(in boardView: BoardView) -> Int
    
    /// Preguntar al Data Source que imagen hay que poner como fondo en una posicion del tablero.
    func backgroundImageName(in boardView: BoardView, atRow row: Int, atColumn column: Int) -> UIImage?
    
    /// Preguntar al Data Source que imagen hay que poner en primer plano en una posicion del tablero.
    func foregroundImageName(in boardView: BoardView, atRow row: Int, atColumn column: Int) -> UIImage?
}


@IBDesignable
class BoardView: UIView {

    /// Data source para obtener los datos del tablero a pintar.
    weak var dataSource: BoardViewDataSource!
    
    @IBInspectable
    var bgColor: UIColor! = UIColor.orange
    
    @IBInspectable
    var borderColor: UIColor! = UIColor.red
    
    
#if TARGET_INTERFACE_BUILDER
    
    var fakeStrongDataSource: FakeDataSource!
    
    /// Data source falso para que lo use el Interface Builder en tiempo de desarrollo.
    class FakeDataSource: BoardViewDataSource {
    
        func numberOfRows(in: BoardView) -> Int {
            return 2
        }
    
        func numberOfColumns(in: BoardView) -> Int {
            return 2
        }
    
        func backgroundImageName(in: BoardView, atRow row: Int, atColumn column: Int) -> UIImage? {
            return nil //UIImage(named: "bg0.png")
        }
    
        func foregroundImageName(in: BoardView, atRow row: Int, atColumn column: Int) -> UIImage? {
            return nil
        }
    }
    
    
    override func prepareForInterfaceBuilder() {
    
        fakeStrongDataSource = FakeDataSource()
        dataSource = fakeStrongDataSource
    }
    
#endif

    
    
    
    override func draw(_ rect: CGRect) {
        updateBoxSize()
        drawBackground()
        drawBlocks()
    }
    
    // Pinta el fondo
    private func drawBackground() {
        
        let lineWidth: CGFloat = 1
        
        // Tamaño del tablero
        let rows = dataSource.numberOfRows(in: self)
        let columns = dataSource.numberOfColumns(in: self)
        
        // Rectangulo de la zona de la view donde pintare en puntos.
        let canvasX = max(0,box2Point(column: 0)-lineWidth)
        let canvasY = max(0,box2Point(row: 0)-lineWidth)
        let canvasWidth  = box2Point(width: columns) + 2*lineWidth
        let canvasHeight = box2Point(height: rows) + 2*lineWidth
        
        let path = UIBezierPath(rect: CGRect(x: canvasX,
                                             y: canvasY,
                                             width: canvasWidth,
                                             height: canvasHeight))
        borderColor.setStroke()
        bgColor.setFill()
        
        path.fill()
        
        path.lineWidth = lineWidth
        path.stroke()
    }
    
    
    // Pinta los bloques añadidos/congelados/depositados en el tablero
    private func drawBlocks() {
        
        // Tamaño del tablero
        let rows = dataSource.numberOfRows(in: self)
        let columns = dataSource.numberOfColumns(in: self)
        
        for r in 0..<rows {
            for c in 0..<columns {
                drawBox(row: r, column: c)
            }
        }
    }
    
    
    // Dibuja el cuadradito de la posicion indicada.
    private func drawBox(row: Int, column: Int) {
        
        let x = box2Point(column: column)
        let y = box2Point(row: row)
        let width = box2Point(width: 1)
        let height = box2Point(height: 1)
        
        let rect = CGRect(x: x, y: y, width: width, height: height)
        
        if let bgImg = dataSource.backgroundImageName(in: self, atRow: row, atColumn: column) {
            bgImg.draw(in: rect)
        }
        
        if let fgImg = dataSource.foregroundImageName(in: self, atRow: row, atColumn: column) {
            fgImg.draw(in: rect)
        }
    }
    
    //-------------
    // Conversiones de coordenadas en cuadraditos, a coordenadas en puntos.
    // Los cuadraditos son los rectangulos que componen los bloques, y que se
    // llaman Box en el codigo.
    // Los puntos son los de la UIView.
    //
    // El area donde se dibuja es la mayor posible manteniendo una relacion de
    // aspecto que hace que los cuadraditos sean cuadrado y ajustando a los bordes
    // de la UIView.
    //-------------
    
    // Cuantos puntos ocupa un cuadradito/Box
    private var boxSize: CGFloat = 0
    
    // Margenes alrededor de la zona de dibujo.
    private let extraMargin: CGFloat = 2
    private var horizontalMargin: CGFloat = 0
    private var verticalMargin: CGFloat = 0
    
    // Actualiza el valor de boxSize
    private func updateBoxSize() {
        
        // Tamaño del tablero
        let rows = dataSource.numberOfRows(in: self)
        let columns = dataSource.numberOfColumns(in: self)
        
        // Tamaño en puntos de la zona de la view donde voy a dibujar.
        let width = bounds.size.width - 2 * extraMargin
        let height = bounds.size.height - 2 * extraMargin
        
        // Tamaño de un cuadradito en puntos
        let boxWidth = width / CGFloat(columns)    // board.columnsCount)
        let boxHeight = height / CGFloat(rows)   // board.rowsCount)
        
        boxSize = min(boxWidth,boxHeight)
        
        horizontalMargin = (width - boxSize * CGFloat(columns)) / 2
        verticalMargin = (height - boxSize * CGFloat(rows)) / 2
    }

    
    // Transforma una coordenada Box representando un fila a puntos.
    private func box2Point(row: Int) -> CGFloat {
        return CGFloat(row) * boxSize + extraMargin + verticalMargin
    }
    
    // Transforma una coordenada Box representando una columna a puntos.
    private func box2Point(column: Int) -> CGFloat {
        return CGFloat(column) * boxSize + extraMargin + horizontalMargin
    }
    
    // Transforma un ancho en unidades Box (columnas) a puntos.
    private func box2Point(width: Int) -> CGFloat {
        return CGFloat(width) * boxSize
    }
    
    // Transforma un alto en unidades Box (filas) a puntos.
    private func box2Point(height: Int) -> CGFloat {
        return CGFloat(height) * boxSize
    }
}






