//
//  ViewController.swift
//  Tetris
//
//  Created by Santiago Pavón Gómez on 9/8/16.
//  Copyright © 2016 UPM. All rights reserved.
//

import UIKit


class ViewController: UIViewController, ScoreDelegate, BoardDelegate, BoardViewDataSource {


    @IBOutlet weak var gamesLabel: UILabel!
    @IBOutlet weak var scoreLabel: UILabel!
    @IBOutlet weak var recordLabel: UILabel!
    
    @IBOutlet weak var boardView: BoardView!
    @IBOutlet weak var nextBlockView: BoardView!
    
    var score = Score()
    var board = Board()
    
    var timer: Timer?
    
    var gameInProgress = false
    

    override func viewDidLoad() {
        super.viewDidLoad()
        
        score.delegate = self
        board.delegate = self
        
        boardView.dataSource = self
        nextBlockView.dataSource = self
        
        startNewGame()

    }

    private func startNewGame() {

        score.newGame()
        board.newGame()
        
        gameInProgress = true
        
        autoMoveDown()
    }
    
    func autoMoveDown() {

        guard gameInProgress else { return }

        board.moveDown(insertNewBlockIfNeeded: true)

        boardView.setNeedsDisplay()
        nextBlockView.setNeedsDisplay()
        
        let interval: TimeInterval = 0.800 - (0.050 * TimeInterval(score.level))
        
        timer = Timer.scheduledTimer(timeInterval: interval,
                                     target: self,
                                     selector: #selector(autoMoveDown),
                                     userInfo: nil,
                                     repeats: false)
    }

    /*
    @IBAction func move(_ sender: UITapGestureRecognizer) {
        
        let width = view.bounds.size.width
        let height = view.bounds.size.height
        
        let pos = sender.location(in: view)
        
        if Float(pos.y / height) > 0.75 {
            moveDown()
        } else if pos.x < width / 2 {
            moveLeft()
        } else if pos.x > width / 2 {
            moveRight()
        }
    }
     */
    
    @IBAction func rotate(_ sender: UIRotationGestureRecognizer) {
   
        if sender.state != .began { return }
        
        if sender.rotation > 0 {
            rotateRight()
        } else {
            rotateLeft()
        }
    }
    
    @IBAction func moveLeft() {
        board.moveLeft()
        boardView.setNeedsDisplay()
    }
    
    
    @IBAction func moveRight() {
        board.moveRight()
        boardView.setNeedsDisplay()

    }
    
    
    @IBAction func moveDown() {
        board.moveDown()
        boardView.setNeedsDisplay()

    }
    
    @IBAction func dropDown(_ sender: UILongPressGestureRecognizer) {
        
        if sender.state != .began { return }

        board.dropDown()
        boardView.setNeedsDisplay()
    }
    
    @IBAction func rotateLeft() {
        board.rotate(toRight: false)
        boardView.setNeedsDisplay()

    }
    
    
    @IBAction func rotateRight() {
        board.rotate(toRight: true)
        boardView.setNeedsDisplay()

    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }

    
    // MARK: - Delegado de Score
    
    func update(games: Int, level: Int, maxRows: Int,  rows: Int) {
        gamesLabel.text = "Partidas: \(games)"
        recordLabel.text = "Record: \(maxRows)"
        scoreLabel.text = "\(rows)"
    }
    
    
    // MARK: - Delegado de Board
    
    
    func rowCompleted() {
        score.rowCompleted()
    }
    
    
    func gameOver() {
        print("Has completado la partida")

        nextBlockView.setNeedsDisplay()
        
        timer?.invalidate()

        gameInProgress = false
        
        // Crear un UIAlertController:
        let alert = UIAlertController(title: "Game Over",
                                      message: "¿Quieres jugar otra partida?",
                                      preferredStyle: UIAlertControllerStyle.alert)

            
        alert.addAction(UIAlertAction(title: "OK", style: .default) {aa in self.startNewGame()})
        
        
        // Presentar el Alert Controller:
        present(alert, animated: true, completion: nil)
        
    }

    // MARK: - Data source del tablero
    
    func numberOfRows(in boardView: BoardView) -> Int {

        switch  boardView {
        case self.boardView:
            return board.rowsCount
        case self.nextBlockView:
            return board.nextBlock?.height ?? 0
        default:
            return 0
        }
    }
    
    func numberOfColumns(in boardView: BoardView) -> Int {
        
        switch  boardView {
        case self.boardView:
            return board.columnsCount
        case self.nextBlockView:
            return board.nextBlock?.width ?? 0
        default:
            return 0
        }
    }
    
    
    func backgroundImageName(in boardView: BoardView, atRow row: Int, atColumn column: Int) -> UIImage? {
        
        switch  boardView {
        case self.boardView:
            if let texture = board.currentTexture(atRow: row, atColumn: column) {
                let imageName = texture.backgroundImageName()
                return cachedImage(name: imageName)
            } else {
               return nil
            }
        case self.nextBlockView:
            guard let block = board.nextBlock,
                block.isSolid(row: row, column: column) else { return nil }
            
            let imageName = block.texture.backgroundImageName()
            return cachedImage(name: imageName)
        default:
            return nil
        }
    }
    
    
    func foregroundImageName(in boardView: BoardView, atRow row: Int, atColumn column: Int) -> UIImage? {
        
        switch  boardView {
        case self.boardView:
            if let texture = board.currentTexture(atRow: row, atColumn: column) {
                let imageName = texture.pokemonImageName()
                return cachedImage(name: imageName)
            } else {
               return nil
            }
        case self.nextBlockView:
            guard let block = board.nextBlock,
                block.isSolid(row: row, column: column) else { return nil }
            
            let imageName = block.texture.pokemonImageName()
            return cachedImage(name: imageName)
        default:
            return nil
        }
    }

    
    // MARK: - Cache de imagenes
    
    var imagesCache = [String:UIImage]()


    private func cachedImage(name imageName: String) -> UIImage? {
        if let image = imagesCache[imageName] {
            return image
        } else if let image = UIImage(named: imageName) {
            imagesCache[imageName] = image
            return image
        }
        return nil
    }
    
}

